#ifndef PBR
#define PBR

#define M_PI 3.14159265358979323846264338327950288

// source: https://github.com/SaschaWillems/Vulkan-glTF-PBR/blob/master/data/shaders/pbr_khr.frag

struct PBRInfo {
    float n_dot_v;
    float roughness;
    float metallic;
    vec3 specular_environmentR0;
    vec3 specular_environmentR90;
    float alpha_roughness;
    vec3 diffuse_color;
    vec3 specular_color;
    vec4 base_color;
    vec3 normal;
    vec3 world_pos;
    vec3 v;
};

#define MANUAL_SRGB

vec4 SRGBtoLINEAR(vec4 srgbIn) {
	#ifdef MANUAL_SRGB
        #ifdef SRGB_FAST_APPROXIMATION
            vec3 linOut = pow(srgbIn.xyz,vec3(2.2));
        #else //SRGB_FAST_APPROXIMATION
            vec3 bLess = step(vec3(0.04045),srgbIn.xyz);
            vec3 linOut = mix( srgbIn.xyz/vec3(12.92), pow((srgbIn.xyz+vec3(0.055))/vec3(1.055),vec3(2.4)), bLess );
        #endif //SRGB_FAST_APPROXIMATION

	    return vec4(linOut,srgbIn.w);;
	#else //MANUAL_SRGB
	    return srgbIn;
	#endif //MANUAL_SRGB
}

// alpha mask
//  0 ... opaque
//  1 ... mask
//  2 ... blend

#define ALPHA_MASK_OPAQUE 0
#define ALPHA_MASK_MASK 1
#define ALPHA_MASK_BLEND 2

bool pbr_color(uint alpha_mask, float alpha_cut_off, inout vec4 result) {
    if (alpha_mask == ALPHA_MASK_OPAQUE) {
        result.a = 1.0;
        return true;
    } else if (alpha_mask == ALPHA_MASK_MASK) {
        if (result.a < alpha_cut_off) {
            result = vec4(0.0);
            return false;
        } else {
            return true;
        }
    } else {
        return true;
    }
}

bool pbr_color(vec4 base_color_factor, sampler2D color_map, vec2 uv, uint alpha_mask, float alpha_cut_off, out vec4 result) {
    result = SRGBtoLINEAR(texture(color_map, uv)) * base_color_factor;

    return pbr_color(alpha_mask, alpha_cut_off, result);
}

bool pbr_color(vec4 base_color_factor, uint alpha_mask, float alpha_cut_off, out vec4 result) {
    result = base_color_factor;

    return pbr_color(alpha_mask, alpha_cut_off, result);
}

const float MIN_ROUGHNESS = 0.04;

void pbr_metallic_roughness(inout float roughness_factor, inout float metallic_factor, sampler2D metallic_roughness_map, vec2 uv) {
    vec4 metallic_roughness_sample = texture(metallic_roughness_map, uv);

    roughness_factor *= metallic_roughness_sample.g;
    metallic_factor *= metallic_roughness_sample.b;
}

void pbr_metallic_roughness(inout float roughness_factor, inout float metallic_factor) {
    roughness_factor = clamp(roughness_factor, MIN_ROUGHNESS, 1.0);
    metallic_factor = clamp(metallic_factor, 0.0, 1.0);
}

vec3 pbr_diffuse(PBRInfo pbr_info) {
    return pbr_info.diffuse_color / M_PI;
}

vec3 specular_reflection(PBRInfo pbr_info, float v_dot_h) {
    return pbr_info.specular_environmentR0 + (pbr_info.specular_environmentR90 - pbr_info.specular_environmentR0) * pow(clamp(1.0 - v_dot_h, 0.0, 1.0), 5.0);
}

float geometric_occlusion(PBRInfo pbr_info, float n_dot_l) {
    float n_dot_v = pbr_info.n_dot_v;
    float r = pbr_info.alpha_roughness;

    float attenuation_l = 2.0 * n_dot_l / ( n_dot_l + sqrt( r * r * (1.0 - r * r) * (n_dot_l * n_dot_l) ) );
    float attenuation_v = 2.0 * n_dot_v / ( n_dot_v + sqrt( r * r * (1.0 - r * r) * (n_dot_v * n_dot_v) ) );

    return attenuation_l * attenuation_v;
}

float microfacet_distribution(PBRInfo pbr_info, float n_dot_h) {
    float roughness_sq = pbr_info.alpha_roughness * pbr_info.alpha_roughness;
    float f = (n_dot_h * roughness_sq - n_dot_h) * n_dot_h + 1.0;

    return roughness_sq / (M_PI * f * f);
}

PBRInfo prepare(vec4 base_color, float roughness, float metallic, vec3 normal, vec3 cam_pos, vec3 world_pos) {
    vec3 f0 = vec3(MIN_ROUGHNESS);

    vec3 diffuse_color = base_color.rgb * (vec3(1.0) - f0);
    diffuse_color *= 1.0 - metallic;

    float alpha_roughness = roughness * roughness;

    vec3 specular_color = mix(f0, base_color.rgb, metallic);

    float reflectance = max(max(specular_color.r, specular_color.g), specular_color.b);

    float reflectance90 = clamp(reflectance * 25.0, 0.0, 1.0);
    vec3 specular_environmentR0 = specular_color.rgb;
    vec3 specular_environmentR90 = vec3(1.0) * reflectance90;

    vec3 v = normalize(cam_pos - world_pos);
    float n_dot_v = clamp(abs(dot(normal, v)), 0.001, 1.0);

    return PBRInfo(
        n_dot_v,
        roughness,
        metallic,
        specular_environmentR0,
        specular_environmentR90,
        alpha_roughness,
        diffuse_color,
        specular_color,
        base_color,
        normal,
        world_pos,
        v
    );
}

vec4 apply_light(PBRInfo pbr_info, vec3 light_pos, vec3 light_color, float light_intensity) {
    vec3 l = normalize(light_pos - pbr_info.world_pos);
    vec3 h = normalize(l + pbr_info.v);

    float n_dot_l = clamp(dot(pbr_info.normal, l), 0.001, 1.0);
    float n_dot_h = clamp(dot(pbr_info.normal, h), 0.0, 1.0);
    float l_dot_h = clamp(dot(l, h), 0.0, 1.0);
    float v_dot_h = clamp(dot(pbr_info.v, h), 0.0, 1.0);

    vec3 F = specular_reflection(pbr_info, v_dot_h);
	float G = geometric_occlusion(pbr_info, n_dot_l);
	float D = microfacet_distribution(pbr_info, n_dot_h);

    vec3 diffuse_contribution = (1.0 - F) * pbr_diffuse(pbr_info);
    vec3 specular_contribution = F * G * D / (4.0 * n_dot_l * pbr_info.n_dot_v);

    vec3 color = light_intensity * n_dot_l * light_color * (diffuse_contribution + specular_contribution);

    return vec4(color, pbr_info.base_color.a);
}

#endif // PBR
